@@ -3,13 +3,12 @@
 ;; Copyright (C) 1995, 1996, 2001 Free Software Foundation, Inc.
 
 ;; Author: Tomoji Kagatani <kagatani@rbc.ncl.omron.co.jp>
-;; Maintainer: Brian D. Carlstrom <bdc@ai.mit.edu>
+;; Maintainer: Simon Josefsson <simon@josefsson.org>
+;; w32 Maintainer: Brian D. Carlstrom <bdc@ai.mit.edu>
 ;; ESMTP support: Simon Leinen <simon@switch.ch>
 ;; Hacked by Mike Taylor, 11th October 1999 to add support for
 ;; automatically appending a domain to RCPT TO: addresses.
 ;; AUTH=LOGIN support: Stephen Cranefield <scranefield@infoscience.otago.ac.nz>
-;; AUTH support: Simon Josefsson <jas@pdc.kth.se>
-;; STARTTLS support: Simon Josefsson <jas@pdc.kth.se>
 ;; Keywords: mail
 
 ;; This file is part of XEmacs.
@@ -29,7 +28,7 @@
 ;; Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 ;; 02111-1307, USA.
 
-;;; Synched up with: FSF 21.0.104
+;;; Synched up with: FSF 21.1
 
 ;;; Commentary:
 
@@ -38,14 +37,11 @@
 ;; Please add these lines in your .emacs(_emacs) or use customize.
 ;;
 ;;(setq send-mail-function 'smtpmail-send-it) ; if you use `mail'
-;;(setq message-send-mail-function 'smtpmail-send-it) ; if you use `message'
+;;(setq message-send-mail-function 'smtpmail-send-it) ; if you use message/Gnus
 ;;(setq smtpmail-default-smtp-server "YOUR SMTP HOST")
 ;;(setq smtpmail-local-domain "YOUR DOMAIN NAME")
 ;;(setq smtpmail-sendto-domain "YOUR DOMAIN NAME")
 ;;(setq smtpmail-debug-info t) ; only to debug problems
-;;(load-library "smtpmail")
-;;(setq smtpmail-code-conv-from nil)
-;;(setq user-full-name "YOUR NAME HERE")
 ;;(setq smtpmail-auth-credentials
 ;;      '(("YOUR SMTP HOST" 25 "username" "password")))
 ;;(setq smtpmail-starttls-credentials
@@ -65,17 +61,22 @@
 ;; See http://www.ietf.org/rfc/rfc2554.txt
 
 ;; Modified by Simon Josefsson <simon@josefsson.org>, 2000-10-07, to support
-;; STARTTLS.  Require external program ftp://ftp.opaopa.org/pub/elisp/starttls-*.tar.gz.
-;; See http://www.ietf.org/rfc/rfc2246.txt, http://www.ietf.org/rfc/rfc2595.txt
+;; STARTTLS.  Requires external program
+;; ftp://ftp.opaopa.org/pub/elisp/starttls-*.tar.gz.
+;; See http://www.ietf.org/rfc/rfc2246.txt, http://www.ietf.org/rfc/rfc2487.txt
 
 ;;; Code:
 
 (require 'sendmail)
 (require 'time-stamp)
-(require 'starttls)
-(require 'rfc2104)
+(autoload 'starttls-open-stream "starttls")
+(autoload 'starttls-negotiate "starttls")
+(autoload 'mail-strip-quoted-names "mail-utils")
+(autoload 'mail-check-safe-charset "mail-utils")
+(autoload 'message-make-date "message")
+(autoload 'message-make-message-id "message")
+(autoload 'rfc2104-hash "rfc2104")
 (require 'base64)
-(eval-when-compile (require 'cl))
 
 ;;;
 (defgroup smtpmail nil
@@ -95,7 +96,8 @@
   :group 'smtpmail)
 
 (defcustom smtpmail-smtp-service 25
-  "*SMTP service port number. smtp or 25 ."
+  "*SMTP service port number.
+The default value would be \"smtp\" or 25 ."
   :type '(choice (integer :tag "Port") (string :tag "Service"))
   :group 'smtpmail)
 
@@ -123,7 +125,15 @@ buffer includes an exchange like:
   :group 'smtpmail)
 
 (defcustom smtpmail-debug-info nil
-  "*smtpmail debug info printout. messages and process buffer."
+  "Whether to print info in buffer *trace of SMTP session to <somewhere>*.
+See also `smtpmail-debug-verb' which determines if the SMTP protocol should
+be verbose as well."
+  :type 'boolean
+  :group 'smtpmail)
+
+(defcustom smtpmail-debug-verb nil
+  "Whether this library sends the SMTP VERB command or not.
+The commands enables verbose information from the SMTP server."
   :type 'boolean
   :group 'smtpmail)
 
@@ -144,21 +154,22 @@ and sent with `smtpmail-send-queued-mail'."
   :type 'directory
   :group 'smtpmail)
 
-(defcustom smtpmail-auth-credentials '(("" 25 "" ""))
-  "*Specify username and password for servers.
+(defcustom smtpmail-auth-credentials '(("" 25 "" nil))
+  "Specify username and password for servers.
 It is a list of four-element lists that contain, in order,
 `servername' (a string), `port' (an integer), `user' (a string) and
-`password' (a string).
+`password' (a string, or nil to query the user when needed).
 If you need to enter a `realm' too, add it to the user string, so that
 it looks like `user@realm'."
   :type '(repeat (list (string  :tag "Server")
 		       (integer :tag "Port")
 		       (string  :tag "Username")
-		       (string  :tag "Password")))
+		       (choice (const :tag "Query when needed" nil)
+			       (string  :tag "Password"))))
   :group 'smtpmail)
 
 (defcustom smtpmail-starttls-credentials '(("" 25 "" ""))
-  "*Specify STARTTLS keys and certificates for servers.
+  "Specify STARTTLS keys and certificates for servers.
 This is a list of four-element list with `servername' (a string),
 `port' (an integer), `key' (a filename) and `certificate' (a filename)."
   :type '(repeat (list (string  :tag "Server")
@@ -196,14 +207,15 @@ This is relative to `smtpmail-queue-dir'.")
 
 ;;;###autoload
 (defun smtpmail-send-it ()
-  (require 'mail-utils)
   (let ((errbuf (if mail-interactive
 		    (generate-new-buffer " smtpmail errors")
 		  0))
 	(tembuf (generate-new-buffer " smtpmail temp"))
 	(case-fold-search nil)
 	delimline
-	(mailbuf (current-buffer)))
+	(mailbuf (current-buffer))
+	;; XEmacs: Don't frob `smtpmail-code-conv-from' here
+	)
     (unwind-protect
 	(save-excursion
 	  (set-buffer tembuf)
@@ -214,11 +226,7 @@ This is relative to `smtpmail-queue-dir'.")
 	  (or (= (preceding-char) ?\n)
 	      (insert ?\n))
 	  ;; Change header-delimiter to be what sendmail expects.
-	  (goto-char (point-min))
-	  (re-search-forward
-	    (concat "^" (regexp-quote mail-header-separator) "\n"))
-	  (replace-match "\n")
-	  (backward-char 1)
+	  (mail-sendmail-undelimit-header)
 	  (setq delimline (point-marker))
 ;;	  (sendmail-synch-aliases)
 	  (if (and mail-aliases (fboundp 'expand-mail-aliases)) ; XEmacs
@@ -286,6 +294,14 @@ This is relative to `smtpmail-queue-dir'.")
 			 (insert ")\n"))
 			((null mail-from-style)
 			 (insert "From: " login "\n")))))
+	    ;; Insert a `Message-Id:' field if there isn't one yet.
+	    (goto-char (point-min))
+	    (unless (re-search-forward "^Message-Id:" delimline t)
+	      (insert "Message-Id: " (message-make-message-id) "\n"))
+	    ;; Insert a `Date:' field if there isn't one yet.
+	    (goto-char (point-min))
+	    (unless (re-search-forward "^Date:" delimline t)
+	      (insert "Date: " (message-make-date) "\n"))
 	    ;; Insert an extra newline if we need it to work around
 	    ;; Sun's bug that swallows newlines.
 	    (goto-char (1+ delimline))
@@ -296,8 +312,7 @@ This is relative to `smtpmail-queue-dir'.")
 	    (if (re-search-forward "^FCC:" delimline t)
 		(mail-do-fcc delimline))
 	    (if mail-interactive
-		(save-excursion
-		  (set-buffer errbuf)
+		(with-current-buffer errbuf
 		  (erase-buffer))))
 	  ;;
 	  ;;
@@ -327,8 +342,7 @@ This is relative to `smtpmail-queue-dir'.")
 		   (buffer-data (create-file-buffer file-data))
 		   (buffer-elisp (create-file-buffer file-elisp))
 		   (buffer-scratch "*queue-mail*"))
-	      (save-excursion
-		(set-buffer buffer-data)
+	      (with-current-buffer buffer-data
 		(erase-buffer)
 		(insert-buffer tembuf)
 		(write-file file-data)
@@ -355,32 +369,30 @@ This is relative to `smtpmail-queue-dir'.")
 (defun smtpmail-send-queued-mail ()
   "Send mail that was queued as a result of setting `smtpmail-queue-mail'."
   (interactive)
-  ;;; Get index, get first mail, send it, get second mail, etc...
-  (let ((buffer-index (find-file-noselect smtpmail-queue-index))
-	(file-msg "")
-	(tembuf nil))
-    (save-excursion
-      (set-buffer buffer-index)
+  (with-temp-buffer
+    ;;; Get index, get first mail, send it, update index, get second
+    ;;; mail, send it, etc...
+    (let ((file-msg ""))
+      (insert-file-contents smtpmail-queue-index)
       (beginning-of-buffer)
       (while (not (eobp))
-	(setq file-msg (buffer-substring (point) (save-excursion
-						   (end-of-line)
-						   (point))))
+	(setq file-msg (buffer-substring (point) (point-at-eol))) ;; XEmacs
 	(load file-msg)
-	(setq tembuf (find-file-noselect file-msg))
-	(if (not (null smtpmail-recipient-address-list))
-	    (if (not (smtpmail-via-smtp smtpmail-recipient-address-list 
-					tembuf))
-		(error "Sending failed; SMTP protocol error"))
-	  (error "Sending failed; no recipients"))  
+	;; Insert the message literally: it is already encoded as per
+	;; the MIME headers, and code conversions might guess the
+	;; encoding wrongly.
+	(with-temp-buffer
+	  (let ((coding-system-for-read 'binary))
+	    (insert-file-contents file-msg))
+	  (if (not (null smtpmail-recipient-address-list))
+	      (if (not (smtpmail-via-smtp smtpmail-recipient-address-list
+					  (current-buffer)))
+		  (error "Sending failed; SMTP protocol error"))
+	    (error "Sending failed; no recipients")))
 	(delete-file file-msg)
 	(delete-file (concat file-msg ".el"))
-	(kill-buffer tembuf)
-	(kill-line 1))      
-      (set-buffer buffer-index)
-      (save-buffer smtpmail-queue-index)
-      (kill-buffer buffer-index)
-      )))
+	(kill-line 1))
+      (write-region (point-min) (point-max) smtpmail-queue-index))))
 
 ;(defun smtpmail-via-smtp (host,port,sender,destination,smtpmail-text-buffer)
 
@@ -389,12 +401,30 @@ This is relative to `smtpmail-queue-dir'.")
       (concat (system-name) "." smtpmail-local-domain)
     (system-name)))
 
+(defsubst smtpmail-cred-server (cred)
+  (nth 0 cred))
+
+(defsubst smtpmail-cred-port (cred)
+  (nth 1 cred))
+
+(defsubst smtpmail-cred-key (cred)
+  (nth 2 cred))
+
+(defsubst smtpmail-cred-user (cred)
+  (nth 2 cred))
+
+(defsubst smtpmail-cred-cert (cred)
+  (nth 3 cred))
+
+(defsubst smtpmail-cred-passwd (cred)
+  (nth 3 cred))
+
 (defun smtpmail-find-credentials (cred server port)
   (catch 'done
     (let ((l cred) el)
       (while (setq el (pop l))
-	(when (and (equal server (nth 0 el))
-		   (equal port (nth 1 el)))
+	(when (and (equal server (smtpmail-cred-server el))
+		   (equal port (smtpmail-cred-port el)))
 	  (throw 'done el))))))
 
 (defun smtpmail-maybe-append-domain (recipient)
@@ -403,6 +433,85 @@ This is relative to `smtpmail-queue-dir'.")
       recipient
     (concat recipient "@" smtpmail-sendto-domain)))
 
+(defun smtpmail-intersection (list1 list2)
+  (let ((result nil))
+    (dolist (el2 list2)
+      (when (memq el2 list1)
+	(push el2 result)))
+    (nreverse result)))
+
+(defun smtpmail-open-stream (process-buffer host port)
+  (let ((cred (smtpmail-find-credentials
+	       smtpmail-starttls-credentials host port)))
+    (if (null (and cred (condition-case ()
+			    (call-process "starttls")
+			  (error nil))))
+	;; The normal case.
+	(open-network-stream "SMTP" process-buffer host port)
+      (let* ((cred-key (smtpmail-cred-key cred))
+	     (cred-cert (smtpmail-cred-cert cred))
+	     (starttls-extra-args
+	      (when (and (stringp cred-key) (stringp cred-cert)
+			 (file-regular-p
+			  (setq cred-key (expand-file-name cred-key)))
+			 (file-regular-p
+			  (setq cred-cert (expand-file-name cred-cert))))
+		(list "--key-file" cred-key "--cert-file" cred-cert))))
+	(starttls-open-stream "SMTP" process-buffer host port)))))
+
+(defun smtpmail-try-auth-methods (process supported-extensions host port)
+  (let* ((mechs (cdr-safe (assoc 'auth supported-extensions)))
+	 (mech (car (smtpmail-intersection smtpmail-auth-supported mechs)))
+	 (cred (smtpmail-find-credentials smtpmail-auth-credentials host port))
+	 (passwd (when cred
+		   (or (smtpmail-cred-passwd cred)
+		       (read-passwd
+			(format "SMTP password for %s:%s: "
+				(smtpmail-cred-server cred)
+				(smtpmail-cred-port cred))))))
+	 ret)
+    (when cred
+      (cond
+       ((eq mech 'cram-md5)
+	(smtpmail-send-command process (format "AUTH %s" mech))
+	(if (or (null (car (setq ret (smtpmail-read-response process))))
+		(not (integerp (car ret)))
+		(>= (car ret) 400))
+	    (throw 'done nil))
+	(when (eq (car ret) 334)
+	  (let* ((challenge (substring (cadr ret) 4))
+		 (decoded (base64-decode-string challenge))
+		 (hash (rfc2104-hash 'md5 64 16 passwd decoded))
+		 (response (concat (smtpmail-cred-user cred) " " hash))
+		 (encoded (base64-encode-string response)))
+	    (smtpmail-send-command process (format "%s" encoded))
+	    (if (or (null (car (setq ret (smtpmail-read-response process))))
+		    (not (integerp (car ret)))
+		    (>= (car ret) 400))
+		(throw 'done nil)))))
+       ((eq mech 'login)
+	(smtpmail-send-command process "AUTH LOGIN")
+	(if (or (null (car (setq ret (smtpmail-read-response process))))
+		(not (integerp (car ret)))
+		(>= (car ret) 400))
+	    (throw 'done nil))
+	(smtpmail-send-command
+	 process (base64-encode-string (smtpmail-cred-user cred)))
+	(if (or (null (car (setq ret (smtpmail-read-response process))))
+		(not (integerp (car ret)))
+		(>= (car ret) 400))
+	    (throw 'done nil))
+	(smtpmail-send-command process (base64-encode-string passwd))
+	(if (or (null (car (setq ret (smtpmail-read-response process))))
+		(not (integerp (car ret)))
+		(>= (car ret) 400))
+	    (throw 'done nil)))
+       (t
+	(error "Mechanism %s not implemented" mech)))
+      ;; Remember the password.
+      (unless (smtpmail-cred-passwd cred)
+	(setcar (cdr (cdr (cdr cred))) passwd)))))
+
 (defun smtpmail-via-smtp (recipient smtpmail-text-buffer)
   (let ((process nil)
 	(host (or smtpmail-smtp-server
@@ -419,30 +528,17 @@ This is relative to `smtpmail-queue-dir'.")
 		(get-buffer-create (format "*trace of SMTP session to %s*" host)))
 
 	  ;; clear the trace buffer of old output
-	  (save-excursion
-	    (set-buffer process-buffer)
+	  (with-current-buffer process-buffer
 	    (erase-buffer))
 
 	  ;; open the connection to the server
-	  (let ((cred (smtpmail-find-credentials smtpmail-starttls-credentials host port)))
-	    (if (and cred (condition-case ()
-			      (call-process "starttls")
-			    (error nil)))
-		(let ((starttls-extra-args
-		       (when (and (string-to-list (nth 2 cred)) (string-to-list (nth 3 cred))
-				  (file-regular-p (expand-file-name (nth 2 cred)))
-				  (file-regular-p (expand-file-name (nth 3 cred))))
-			 (list "--key-file" (expand-file-name (nth 2 cred))
-			       "--cert-file" (expand-file-name (nth 3 cred))))))
-		  (setq process (starttls-open-stream "SMTP" process-buffer host port)))
-	      (setq process (open-network-stream "SMTP" process-buffer host port))))
+	  (setq process (smtpmail-open-stream process-buffer host port))
 	  (and (null process) (throw 'done nil))
 
 	  ;; set the send-filter
 	  (set-process-filter process 'smtpmail-process-filter)
 
-	  (save-excursion
-	    (set-buffer process-buffer)
+	  (with-current-buffer process-buffer
 	    ;; XEmacs change
 	    (if (or (featurep 'mule) (featurep 'file-coding))
 		(set-process-coding-system process 'no-conversion-unix 'no-conversion-unix))
@@ -459,92 +555,56 @@ This is relative to `smtpmail-queue-dir'.")
 	    (let ((do-ehlo t)
 		  (do-starttls t))
 	      (while do-ehlo
-		;; EHLO
-		(smtpmail-send-command process (format "EHLO %s" (smtpmail-fqdn)))
+	    ;; EHLO
+	    (smtpmail-send-command process (format "EHLO %s" (smtpmail-fqdn)))
 
-		(if (or (null (car (setq response-code (smtpmail-read-response process))))
-			(not (integerp (car response-code)))
-			(>= (car response-code) 400))
-		    (progn
-		      ;; HELO
-		      (smtpmail-send-command process (format "HELO %s" (smtpmail-fqdn)))
+	    (if (or (null (car (setq response-code
+				     (smtpmail-read-response process))))
+		    (not (integerp (car response-code)))
+		    (>= (car response-code) 400))
+		(progn
+		  ;; HELO
+		  (smtpmail-send-command
+		   process (format "HELO %s" (smtpmail-fqdn)))
 
-		      (if (or (null (car (setq response-code (smtpmail-read-response process))))
-			      (not (integerp (car response-code)))
-			      (>= (car response-code) 400))
-			  (throw 'done nil)))
-		  (let ((extension-lines (cdr (cdr response-code))))
-		    (while extension-lines
-		      (let ((name (mapcar 'intern (mapcar 'downcase (split-string (substring (car extension-lines) 4) "[ ]")))))
-			(and (eq (length name) 1)
-			     (setq name (car name)))
-			(and name
-			     (cond ((memq (if (consp name) (car name) name)
-					  '(verb xvrb 8bitmime onex xone
-						 expn size dsn etrn 
-						 enhancedstatuscodes
-						 help xusr
-						 auth=login auth starttls))
-				    (setq supported-extensions
-					  (cons name supported-extensions)))
-				   (smtpmail-warn-about-unknown-extensions
-				    (message "Unknown extension %s" name)))))
-		      (setq extension-lines (cdr extension-lines)))))
-	  
-		(if (and do-starttls
-			 (smtpmail-find-credentials smtpmail-starttls-credentials host port)
-			 (member 'starttls supported-extensions)
-			 (process-id process))
-		    (progn
-		      (smtpmail-send-command process (format "STARTTLS"))
-		      (if (or (null (car (setq response-code (smtpmail-read-response process))))
-			      (not (integerp (car response-code)))
-			      (>= (car response-code) 400))
-			  (throw 'done nil))
-		      (starttls-negotiate process)
-		      (setq do-starttls nil))
-		  (setq do-ehlo nil))))
+		  (if (or (null (car (setq response-code
+					   (smtpmail-read-response process))))
+			  (not (integerp (car response-code)))
+			  (>= (car response-code) 400))
+		      (throw 'done nil)))
+	      (dolist (line (cdr (cdr response-code)))
+		(let ((name (mapcar (lambda (s) (intern (downcase s)))
+				    (split-string (substring line 4) "[ ]"))))
+		  (and (eq (length name) 1)
+		       (setq name (car name)))
+		    (and name
+		       (cond ((memq (if (consp name) (car name) name)
+				    '(verb xvrb 8bitmime onex xone
+						  expn size dsn etrn
+				      enhancedstatuscodes
+				      help xusr
+				      auth=login auth starttls))
+				(setq supported-extensions
+				      (cons name supported-extensions)))
+			       (smtpmail-warn-about-unknown-extensions
+			      (message "Unknown extension %s" name)))))))
+
+	    (if (and do-starttls
+		     (smtpmail-find-credentials smtpmail-starttls-credentials host port)
+		     (member 'starttls supported-extensions)
+		     (process-id process))
+		(progn
+		  (smtpmail-send-command process (format "STARTTLS"))
+		  (if (or (null (car (setq response-code (smtpmail-read-response process))))
+			  (not (integerp (car response-code)))
+			  (>= (car response-code) 400))
+		      (throw 'done nil))
+		  (starttls-negotiate process)
+		  (setq do-starttls nil))
+	      (setq do-ehlo nil))))
+	    
+	    (smtpmail-try-auth-methods process supported-extensions host port)
 
-	    (let* ((mechs (cdr-safe (assoc 'auth supported-extensions)))
-		   (mech (car (intersection smtpmail-auth-supported mechs)))
-		   (cred (smtpmail-find-credentials smtpmail-auth-credentials host port)))
-	      (when cred
-		(cond ((eq mech 'cram-md5)
-		       (smtpmail-send-command process (format "AUTH %s" mech))
-		       (if (or (null (car (setq response-code (smtpmail-read-response process))))
-			       (not (integerp (car response-code)))
-			       (>= (car response-code) 400))
-			   (throw 'done nil))
-		       (when (eq (car response-code) 334)
-			 (let* ((challenge (substring (cadr response-code) 4))
-				(decoded (base64-decode-string challenge))
-				(hash (rfc2104-hash 'md5 64 16 (nth 3 cred) decoded))
-				(response (concat (nth 2 cred) " " hash))
-				(encoded (base64-encode-string response)))
-			   (smtpmail-send-command process (format "%s" encoded))
-			   (if (or (null (car (setq response-code (smtpmail-read-response process))))
-				   (not (integerp (car response-code)))
-				   (>= (car response-code) 400))
-			       (throw 'done nil)))))
-		      ((eq mech 'login)
-		       (smtpmail-send-command process "AUTH LOGIN")
-		       (if (or (null (car (setq response-code (smtpmail-read-response process))))
-			       (not (integerp (car response-code)))
-			       (>= (car response-code) 400))
-			   (throw 'done nil))
-		       (smtpmail-send-command process (base64-encode-string (nth 2 cred)))
-		       (if (or (null (car (setq response-code (smtpmail-read-response process))))
-			       (not (integerp (car response-code)))
-			       (>= (car response-code) 400))
-			   (throw 'done nil))
-		       (smtpmail-send-command process (base64-encode-string (nth 3 cred)))
-		       (if (or (null (car (setq response-code (smtpmail-read-response process))))
-			       (not (integerp (car response-code)))
-			       (>= (car response-code) 400))
-			   (throw 'done nil)))
-		      (t
-		       (error "Mechanism %s not implemented" mech)))))
-		
 	    (if (or (member 'onex supported-extensions)
 		    (member 'xone supported-extensions))
 		(progn
@@ -554,7 +614,7 @@ This is relative to `smtpmail-queue-dir'.")
 			  (>= (car response-code) 400))
 		      (throw 'done nil))))
 
-	    (if (and smtpmail-debug-info
+	    (if (and smtpmail-debug-verb
 		     (or (member 'verb supported-extensions)
 			 (member 'xvrb supported-extensions)))
 		(progn
@@ -577,8 +637,7 @@ This is relative to `smtpmail-queue-dir'.")
 		   (if (or (member 'size supported-extensions)
 			   (assoc 'size supported-extensions))
 		       (format " SIZE=%d"
-			       (save-excursion
-				 (set-buffer smtpmail-text-buffer)
+			       (with-current-buffer smtpmail-text-buffer
 				 ;; size estimate:
 				 (+ (- (point-max) (point-min))
 				    ;; Add one byte for each change-of-line
@@ -609,8 +668,8 @@ This is relative to `smtpmail-queue-dir'.")
 		     "")))
 ;	      (smtpmail-send-command process (format "MAIL FROM:%s@%s" (user-login-name) (smtpmail-fqdn)))
 	      (smtpmail-send-command process (format "MAIL FROM: <%s>%s%s"
-						     ;; XEmacs:
 						     (or mail-envelope-from
+							 ;; XEmacs:
 							 (user-mail-address))
 						     size-part
 						     body-part))
@@ -662,8 +721,7 @@ This is relative to `smtpmail-queue-dir'.")
 ;		 (throw 'done nil))
 	    t ))
       (if process
-	  (save-excursion
-	    (set-buffer (process-buffer process))
+	  (with-current-buffer (process-buffer process)
 	    (smtpmail-send-command process "QUIT")
 	    (smtpmail-read-response process)
 
@@ -676,8 +734,7 @@ This is relative to `smtpmail-queue-dir'.")
 
 
 (defun smtpmail-process-filter (process output)
-  (save-excursion
-    (set-buffer (process-buffer process))
+  (with-current-buffer (process-buffer process)
     (goto-char (point-max))
     (insert output)))
 
@@ -749,7 +806,7 @@ This is relative to `smtpmail-queue-dir'.")
   (if (and (or (featurep 'mule) (featurep 'file-coding))
 	   (not (null smtpmail-code-conv-from)))
       (setq data (encode-coding-string data smtpmail-code-conv-from)))
-	
+
   (if smtpmail-debug-info
       (insert data "\r\n"))
 
@@ -768,13 +825,11 @@ This is relative to `smtpmail-queue-dir'.")
        this-line
        this-line-end)
 
-    (save-excursion
-      (set-buffer buffer)
+    (with-current-buffer buffer
       (goto-char (point-min)))
 
     (while data-continue
-      (save-excursion
-	(set-buffer buffer)
+      (with-current-buffer buffer
 	(beginning-of-line)
 	(setq this-line (point))
 	(end-of-line)
@@ -792,11 +847,9 @@ This is relative to `smtpmail-queue-dir'.")
 
 (defun smtpmail-deduce-address-list (smtpmail-text-buffer header-start header-end)
   "Get address list suitable for smtp RCPT TO: <address>."
-  (require 'mail-utils)  ;; pick up mail-strip-quoted-names
-    
   (unwind-protect
-      (save-excursion
-	(set-buffer smtpmail-address-buffer) (erase-buffer)
+      (with-current-buffer smtpmail-address-buffer
+	(erase-buffer)
 	(let
 	    ((case-fold-search t)
 	     (simple-address-list "")
@@ -807,9 +860,11 @@ This is relative to `smtpmail-queue-dir'.")
 	  (goto-char (point-min))
 	  ;; RESENT-* fields should stop processing of regular fields.
 	  (save-excursion
-	    (if (re-search-forward "^Resent-\\(to\\|cc\\|bcc\\):" header-end t)
-		(setq addr-regexp "^Resent-\\(to\\|cc\\|bcc\\):")
-	      (setq addr-regexp  "^\\(To:\\|Cc:\\|Bcc:\\)")))
+	    (setq addr-regexp
+		  (if (re-search-forward "^Resent-\\(to\\|cc\\|bcc\\):"
+					 header-end t)
+		      "^Resent-\\(to\\|cc\\|bcc\\):"
+		    "^\\(To:\\|Cc:\\|Bcc:\\)")))
 
 	  (while (re-search-forward addr-regexp header-end t)
 	    (replace-match "")
@@ -824,9 +879,7 @@ This is relative to `smtpmail-queue-dir'.")
 			  (mail-strip-quoted-names (buffer-substring this-line this-line-end))))
 	    )
 	  (erase-buffer)
-	  (insert-string " ")
-	  (insert-string simple-address-list)
-	  (insert-string "\n")
+	  (insert " " simple-address-list "\n")
 	  (subst-char-in-region (point-min) (point-max) 10 ?  t);; newline --> blank
 	  (subst-char-in-region (point-min) (point-max) ?, ?  t);; comma   --> blank
 	  (subst-char-in-region (point-min) (point-max)  9 ?  t);; tab     --> blank
